﻿SpeakinSpell = LibStub("AceAddon-3.0"):NewAddon("SpeakinSpell", "AceConsole-3.0", "AceEvent-3.0")

-------------------------------------------------------------------------------
-- CONSTANTS
-------------------------------------------------------------------------------


-- load localized text including default speeches
local L = LibStub("AceLocale-3.0"):GetLocale("SpeakinSpell", false)


SpeakinSpell.CURRENT_VERSION = "3.0.3.16"

local MAX_MESSAGES_PER_SPELL = 50
local URL = "http://www.wowace.com/projects/speakinspell/"

local FREQUENCY_DEFAULT = 1.00 -- means 100%
local COOLDOWN_DEFAULT = 0 -- default of 0 seconds = no cooldown


-- create a reverse lookup table to un-localize channel names
-- so we can return a channel name that is compatible with SendChatMessage()
local CHANNEL_TABLE = {
	[L["SAY"]]			= "SAY",
	[L["PARTY"]]		= "PARTY",
	[L["RAID"]]			= "RAID",
	[L["EMOTE"]]		= "EMOTE",
	[L["BATTLEGROUND"]]	= "BATTLEGROUND",
	[L["YELL"]]			= "YELL",
	[L["RAID_WARNING"]]	= "RAID_WARNING",
}


-------------------------------------------------------------------------------
-- CHANNEL RULES
-------------------------------------------------------------------------------


--NOTE: these need to be UN-localized chat channel names for use in SendChatMessage
local DEFAULT_CHANNEL_RULES = {
	BG		= nil,		--L["BATTLEGROUND"],
	Raid	= nil,		--L["RAID"],
	Party	= "SAY",	--L["PARTY"]
	Solo	= "SAY",
}


-------------------------------------------------------------------------------
-- DEBUG MESSAGES
-------------------------------------------------------------------------------


function SpeakinSpell:Print(message)
	-- Overriding AceConsole-3.0:Print to use localized addon name instead of tostring(self)
	local text = "|cff33ff99"..--[[tostring( self )]] L["SpeakinSpell"].."|r: " .. message
	DEFAULT_CHAT_FRAME:AddMessage( text )
end


function SpeakinSpell:DebugMsg(message)
	--self:Print(message) -- comment this out to suppress debug messages
end


function SpeakinSpell:DebugMsgDumpString(name, value)
	if value and type(value)=="string" then
		self:DebugMsg(name .. ":" .. value)
	end
end


function SpeakinSpell:DebugMsgDumpBool(name, value)
	if value then
		self:DebugMsg(name.."=true")
	else
		self:DebugMsg(name.."=false")
	end
end


function SpeakinSpell:DebugMsgDumpTable(Table, name)
	for k,v in pairs(Table) do
		if type(v) == "table" then
			self:DebugMsgDumpTable(v, name.."."..k)
		else
			self:DebugMsg( name.."["..k.."]".."="..tostring(v) )
		end
	end
end


-------------------------------------------------------------------------------
-- POPUP DIALOGS
-------------------------------------------------------------------------------


StaticPopupDialogs["NOT_A_NEW_SPELL"] = {
  text = L["POPUPTEXT_NOT_A_NEW_SPELL"], -- will be changed to show the spell name
  button1 = L["OK"],
  -- button2 = L["No"],
  OnAccept = function() end,
  timeout = 0,
  whileDead = 1,
  hideOnEscape = 1
};


-------------------------------------------------------------------------------
-- UTILITY FUNCTIONS
-------------------------------------------------------------------------------


-- create a deep copy of the table and return it
function SpeakinSpell:CopyTable(SourceTable)
	local dest = {} -- make a new table in 'dest'
	for k,v in pairs(SourceTable) do
		-- dig into embedded tables to copy those too, not just create references
		if type(v) == "table" then
			dest[k] = self:CopyTable(v)
		else
			dest[k] = v
		end
	end
	return dest
end


function SpeakinSpell:SelectFirstSpellIfNoneSelected()
	self:DebugMsg("SelectFirstSpellIfNoneSelected()")
	
	-- allow the current selection to remain if it is valid
	if	self.RuntimeData.CurrentSpellSelected and 
		self.RuntimeData.CurrentSpellSelected ~= "" and 
		SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected] then --NOTE: go straight to SpellTable, bypassing GetSpellTableEntry which checks name and rank separately
		-- the current selection is valid so keep it
		return
	end
	
	-- select the first thing in the list, if possible
	for k,v in pairs(SpeakinSpellSavedData.SpellTable) do
		self.RuntimeData.CurrentSpellSelected = k
		return
	end
	
	-- our spell list is empty
	self.RuntimeData.CurrentSpellSelected = nil
end



function SpeakinSpell:DeleteSpell(spellnamekey)
	self:DebugMsg("DeleteSpell( " ..spellnamekey.. " )")
	
	-- delete the current selection from the SavedData and RuntimeData tables
	SpeakinSpellSavedData.SpellTable[spellnamekey] = nil
	
	if not self.RuntimeData then
		-- we're still initializing, so the rest of this stuff doesn't matter
		return
	end
	
	self.RuntimeData.LastMessage[spellnamekey] = nil
	self.RuntimeData.SpellList[spellnamekey] = nil
	
	self:SelectFirstSpellIfNoneSelected() -- make sure we still have a valid spell selection
	self:ShowHideOptionsGUIControls()
end


function SpeakinSpell:CreateNewSpellMessageTable(spellname)
	self:DebugMsg("CreateNewSpellMessageTable()")
	
	spellname_lowercase = string.lower(spellname)
	-- enter the new spell into the spell tabe
	SpeakinSpellSavedData.SpellTable[spellname_lowercase] = {
		Channels = self:CopyTable(DEFAULT_CHANNEL_RULES),
		WhisperTarget = false,
		Frequency = FREQUENCY_DEFAULT,
		Messages = {
			--[1] = string.gsub(L["DEFAULT_NEW_SPELL_MESSAGE"], L["<spellname>"], spellname), --"<player> is casting <spellname> on <target>"
			[1] = L["DEFAULT_NEW_SPELL_MESSAGE"], -- allow <spellname> to remain a variable in the message, as a demonstration
		},
	}
	
	-- enter the new spell name into the list for the options GUI
	self.RuntimeData.SpellList[spellname_lowercase] = spellname_lowercase
end


-------------------------------------------------------------------------------
-- OPTIONS GUI FUNCTIONS
-------------------------------------------------------------------------------


function SpeakinSpell:OnClickNew()
	self:DebugMsg("OnClickNew()")
	
	-- hide everything else and show the new spell text box
	self:EnableNewSpellMode(true)
end


function SpeakinSpell:OnClickCancelNewSpell()
	self:DebugMsg("OnClickCancelNewSpell()")
	self:EnableNewSpellMode(false)
end


function SpeakinSpell:EnableNewSpellMode(enable)
	self:DebugMsg("EnableNewSpellMode()")
	-- turn on EnterNewSpellMode and update show/hide state of all options GUI controls
	self.RuntimeData.EnterNewSpellMode = enable
	self:ShowHideOptionsGUIControls()
end


function SpeakinSpell:OnSetNewSpellName(spellname)
	-- check for user cancelation by empty string
	if not spellname or spellname == "" then
		self:DebugMsg("OnSetNewSpellName() - spell name is empty - user cancelation")
		self:EnableNewSpellMode(false) -- exit "new spell" mode
		return
	end
	
	-- force the spell name to lower case because that's what we use for table keys
	spellname = string.lower(spellname)
	
	-- make sure the new spell name is not already in the list
	-- NOTE: bypass this... SpelltableEntry = self:GetSpellTableEntry(spellname, rank)
	--		go straight to SpeakinSpellSavedData.SpellTable[name]
	--		because spellname here might include the rank, but GetSpellTableEntry wants the name and rank separately
	if SpeakinSpellSavedData.SpellTable[spellname] then
		-- tell user that spell is already in the list
		StaticPopupDialogs["NOT_A_NEW_SPELL"].text = string.gsub(L["POPUPTEXT_NOT_A_NEW_SPELL"], L["<spellname>"], spellname)
		StaticPopup_Show("NOT_A_NEW_SPELL");
		self:DebugMsg("OnSetNewSpellName() - the spell is already in the list")
		self.RuntimeData.CurrentSpellSelected = spellname
		self:EnableNewSpellMode(false) -- exit "new spell" mode
		return
	end
	
	-- add the spell to the message table and related data structures
	self:CreateNewSpellMessageTable(spellname)
	
	-- set the spell selection to the new spell
	self.RuntimeData.CurrentSpellSelected = spellname
	
	-- exit "new spell" mode
	self:EnableNewSpellMode(false)
end


function SpeakinSpell:OnClickDelete()
	self:DebugMsg("OnClickDelete()")
	
	-- make sure we have a valid spell selection
	if not self.RuntimeData.CurrentSpellSelected or self.RuntimeData.CurrentSpellSelected == "" then
		self:DebugMsg("OnClickDelete() - nothing is selected to delete")
		return
	end
	
	-- delete the current selection from the SavedData table
	self:DeleteSpell(self.RuntimeData.CurrentSpellSelected)
end


function SpeakinSpell:OnGetSpellSelect()
	self:DebugMsg("OnGetSpell()")
	
	-- create a side-effect to update the shown/hidden state of all options GUI controls
	self:ShowHideOptionsGUIControls()
	
	-- return the current spell selection
	return self.RuntimeData.CurrentSpellSelected
end


function SpeakinSpell:OnSetSpellSelect(value)
	self:DebugMsg("OnSetSpell( " .. value .. " )")
	-- update the current selection
	self.RuntimeData.CurrentSpellSelected = value
end


function SpeakinSpell:OnSetChannelSelect(key, value)
	self:DebugMsg("OnSetChannelSelect( " .. value .. " )")
	if not self.RuntimeData.CurrentSpellSelected or self.RuntimeData.CurrentSpellSelected == "" then
		self:DebugMsg("OnSetChannelSelect() - no spell is selected")
		return
	end
	-- store the new channel selection
	if value == L["Silent"] then
		SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected].Channels[key] = nil
	else
		SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected].Channels[key] = CHANNEL_TABLE[value]
	end
end


function SpeakinSpell:OnGetChannelSelect(key)
	self:DebugMsg("OnGetChannelSelect()")
	
	-- make sure we have a valid spell selection
	if not self.RuntimeData.CurrentSpellSelected or self.RuntimeData.CurrentSpellSelected == "" then
		self:DebugMsg( "OnGetChannelSelect - no spell selected" )
		-- return a default value
		if DEFAULT_CHANNEL_RULE[key] then
			return DEFAULT_CHANNEL_RULE[key]
		else
			return L["Silent"] -- we store nil values but show "Silent" text
		end
	end
	
	-- look up the stored settings for the channel selection rules for this spell
	local Channels = SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected].Channels
	if not Channels then
		self:DebugMsg( "OnGetChannelSelect - Channels table is missing, using defaults" )
		SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected].Channels = self.CopyTable(DEFAULT_CHANNEL_RULE)
		Channels = SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected].Channels
	end
	
	-- looks like we have a channel selection table, so return it
	if Channels[key] then
		return L[Channels[key]]
	else
		return L["Silent"] -- we store nil values but show "Silent" text
	end
end


function SpeakinSpell:OnGetWhisperTarget()
	self:DebugMsg("OnGetWhisperTarget()")
	
	-- make sure we have a valid spell selection
	if not self.RuntimeData.CurrentSpellSelected or self.RuntimeData.CurrentSpellSelected == "" then
		self:DebugMsg( "OnGetWhisperTarget - no spell selected" )
		return false
	end
	
	-- make sure we have a spell table entry
	local SpellTableEntry = SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected]
	if not SpellTableEntry then
		self:DebugMsg( "OnGetWhisperTarget - ERROR: SpellTableEntry does not exist" )
		return false
	end
	
	-- make sure the WhisperTarget setting exists in the Spelltable
	if SpellTableEntry.WhisperTarget == nil then
		SpellTableEntry.WhisperTarget = false
	end

	-- return the WhisperTarget setting, which should be true or false
	return SpellTableEntry.WhisperTarget
end


function SpeakinSpell:OnSetWhisperTarget(value)
	self:DebugMsg("OnSetWhisperTarget( " ..tostring(value) .. " )")
	
	-- make sure we have a valid spell selection
	if not self.RuntimeData.CurrentSpellSelected or self.RuntimeData.CurrentSpellSelected == "" then
		self:DebugMsg( "OnSetWhisperTarget - no spell selected" )
		return
	end
	
	-- make sure we have a spell table entry
	local SpellTableEntry = SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected]
	if not SpellTableEntry then
		self:DebugMsg( "OnSetWhisperTarget - ERROR: SpellTableEntry does not exist" )
		return
	end
	
	-- store the new setting
	SpellTableEntry.WhisperTarget = value
end


function SpeakinSpell:OnGetMessage(index)
	--self:DebugMsg( "OnGetMessage( " .. index .. " )" )
	
	if not self.RuntimeData.CurrentSpellSelected or self.RuntimeData.CurrentSpellSelected == "" then
		self:DebugMsg( "OnGetMessage( " .. index .. " ) - no selection" )
		return ""
	end
	
	if not SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected].Messages then
		self:DebugMsg( "OnGetMessage( " .. index .. " ) - no message table for current selection: " .. self.RuntimeData.CurrentSpellSelected )
		return ""
	end
	
	local msg = SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected].Messages[index]
	if not msg then
		msg = ""
	end
	
	--self:DebugMsg( "OnGetMessage( " .. index .. " ) = " .. msg )
	return msg
end


function SpeakinSpell:OnSetMessage(index, value)
	if not value then
		value = ""
	end

	self:DebugMsg("OnSetMessage( " .. index .. "= " .. value .." )")
	
	-- make sure we have a selected spell
	if not self.RuntimeData.CurrentSpellSelected or self.RuntimeData.CurrentSpellSelected == "" then
		self:DebugMsg( "OnSetMessage( " .. index .. " ) - no selection" )
		return
	end
	
	-- make sure we have a message table for the currently selected spell
	if not SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected].Messages then
		SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected].Messages = {}
	end
	
	-- store the new value into the message table
	SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected].Messages[index] = value
	self:DebugMsg( "Message " .. index .. " for spell " .. self.RuntimeData.CurrentSpellSelected .. " set to: " .. SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected].Messages[index] )
	
	-- remove holes from the message table
	-- if the user has just deleted one of the entries, move everything else up to fill in the gap
	if value == "" then
		local Messages = SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected].Messages -- create reference for shortcut notation
		for i=index+1,MAX_MESSAGES_PER_SPELL,1 do
			Messages[i-1] = Messages[i]
		end
		Messages[MAX_MESSAGES_PER_SPELL] = nil
		-- destroy empty strings from the table
		for i=MAX_MESSAGES_PER_SPELL-1,1,-1 do
			if Messages[i] == "" then
				Messages[i] = nil
			else
				break --the rest should be valid non-empty strings
			end
		end
		-- hide the superfluous input boxes
		self:ShowHideOptionsGUIControls()
	end
end


function SpeakinSpell:OnGetShowSpellEventsToggle()
	-- if SpeakinSpellSavedData.ShowSpellEvents is nil, translate to false
	if SpeakinSpellSavedData.ShowSpellEvents then
		return true
	else
		return false
	end
end


function SpeakinSpell:OnSetShowSpellEventsToggle(value)
	SpeakinSpellSavedData.ShowSpellEvents = value
end


function SpeakinSpell:OnGetFrequencySelect()
	-- NOTE: if we are running on data from an older version of SpeakinSpell
	--		it is possible that the FrequencySelect value is nil
	--		in that case, use the default value of 100%
	SpellTableEntry = SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected]
	if SpellTableEntry.Frequency == nil then
		SpellTableEntry.Frequency = FREQUENCY_DEFAULT
	end
	
	return SpellTableEntry.Frequency
end


function SpeakinSpell:OnSetFrequencySelect(value)	
	-- SpellTableEntry.Frequency stores a number from 0.01 to 1.00 representing a percentage
	SpellTableEntry = SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected]
	SpellTableEntry.Frequency = value
end


function SpeakinSpell:OnGetCooldownSlider()
	-- NOTE: if we are running on data from an older version of SpeakinSpell
	--		it is possible that the Cooldown value is nil
	--		in that case, use the default value of 0 seconds
	SpellTableEntry = SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected]
	if SpellTableEntry.Cooldown == nil then
		SpellTableEntry.Cooldown = COOLDOWN_DEFAULT
	end
	return SpellTableEntry.Cooldown
end


function SpeakinSpell:OnSetCooldownSlider(value)
	SpellTableEntry = SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected]
	SpellTableEntry.Cooldown = value
end


function SpeakinSpell:ShowHideOptionsGUIControls()
	self:DebugMsg("ShowHideOptionsGUIControls()")
	self:DebugMsgDumpTable(self.RuntimeData.SpellList, "self.RuntimeData.SpellList")
	
	-- safety check in case this function is called early
	if not self.OptionsGUI then
		self:DebugMsg("ShowHideOptionsGUIControls() - called prematurely, abort")
		return
	end
	
	-- establish some conditions that control visibility of several controls
	-- NOTE: for some unknown reason #(self.RuntimeData.SpellList) always returns 0 even if the SpellList is not empty
	local SpellListIsEmpty = true
	for k,v in pairs(self.RuntimeData.SpellList) do -- DebugMsgDumpTable works as intended, unlike #(SpellList), and it uses a for loop like this, so this should work
		SpellListIsEmpty = false
		break
	end
	local SpellListIsNonEmpty = not SpellListIsEmpty

	local EnterNewSpellMode = self.RuntimeData.EnterNewSpellMode
	local HaveSelectedSpell = (self.RuntimeData and self.RuntimeData.CurrentSpellSelected and self.RuntimeData.CurrentSpellSelected ~= "")
	
	--[[
	self:DebugMsgDumpBool("HaveSelectedSpell",	HaveSelectedSpell)
	self:DebugMsgDumpBool("SpellListIsEmpty",	SpellListIsEmpty)
	self:DebugMsgDumpBool("SpellListIsNonEmpty",SpellListIsNonEmpty)
	self:DebugMsgDumpBool("EnterNewSpellMode",	EnterNewSpellMode)
	--]]
	
	-- This table defines the condition under which each options GUI control is shown
	-- if not listed, then the control is always shown
	local ShowControlsWhen = {
		-- controls used for "new spell" mode
		["NewSpellName"]		= EnterNewSpellMode,
		["CancelNewSpellButton"]= EnterNewSpellMode,
		["NewSpellModeInstruct"]= EnterNewSpellMode,
		-- controls that are hidden for "new spell" mode
		["CaptionNormalMode"]	= (not EnterNewSpellMode),
		["ShowSpellEventsToggle"]=(not EnterNewSpellMode),
		["NewSpellButton"]		= (not EnterNewSpellMode),
		["DeleteSpellButton"]	= (not EnterNewSpellMode) and SpellListIsNonEmpty and HaveSelectedSpell,
		["SpellSelect"]			= (not EnterNewSpellMode) and SpellListIsNonEmpty,
		["ChannelSelectCaption"]= (not EnterNewSpellMode) and SpellListIsNonEmpty and HaveSelectedSpell,
		["ChannelSelectBG"]		= (not EnterNewSpellMode) and SpellListIsNonEmpty and HaveSelectedSpell,
		["ChannelSelectRaid"]	= (not EnterNewSpellMode) and SpellListIsNonEmpty and HaveSelectedSpell,
		["ChannelSelectParty"]	= (not EnterNewSpellMode) and SpellListIsNonEmpty and HaveSelectedSpell,
		["ChannelSelectSolo"]	= (not EnterNewSpellMode) and SpellListIsNonEmpty and HaveSelectedSpell,
		["WhisperTargetToggle"]	= (not EnterNewSpellMode) and SpellListIsNonEmpty and HaveSelectedSpell,
		["FrequencySelect"]		= (not EnterNewSpellMode) and SpellListIsNonEmpty and HaveSelectedSpell,
		["CooldownSlider"]		= (not EnterNewSpellMode) and SpellListIsNonEmpty and HaveSelectedSpell,
		-- instructions for setup of first spell, if the spell list is empty
		["AddFirstSpellInstruct"] = SpellListIsEmpty,
	}

	-- determine how many Spell Message N input boxes are shown
	-- typically we'll show at least 1 speech message input box if we have a spell selected
	-- we want to show 1 more than the number of speeches currently configured
	-- unless we don't have a spell selected or we're in EnterNewSpell mode
	local NumSpellMessageInputsShown = 0	
	if HaveSelectedSpell and SpellListIsNonEmpty and (not EnterNewSpellMode) then
		NumSpellMessageInputsShown = 1 + #(SpeakinSpellSavedData.SpellTable[self.RuntimeData.CurrentSpellSelected].Messages)
		self:DebugMsg("NumSpellMessageInputsShown="..tostring(NumSpellMessageInputsShown))
	end
	
	-- show SpellMessageN input box if N <= NumSpellMessageInputsShown (numbering starts at 1)
	for i=1,MAX_MESSAGES_PER_SPELL,1 do
		ShowControlsWhen["SpellMessage"..i] = (i <= NumSpellMessageInputsShown)
	end

	-- toggle the shown/hidden state of all applicable controls based on the ShowControlsWhen table
	for name,show in pairs(ShowControlsWhen) do
		self.OptionsGUI.args.General.args[name].guiHidden = not show
	end
end


function SpeakinSpell:SetupOptionsGUI()
	self:DebugMsg("SetupOptionsGUI()")
	--self:DebugMsgDumpTable(self.RuntimeData.SpellList, "self.RuntimeData.SpellList")

	-- Options table for the GUI
	self.OptionsGUI = {
		type = "group",
		name = L["SpeakinSpell"],
		-- default = function() self:ResetToDefaults() end, -- NOTE: we can't add this here or Ace will complain - we'll add it to the frame object at the end
		args = {
			General = {
				order = 1,
				type = "group",
				name = L["General Settings"],
				desc = L["General Settings TOOLTIP"],
				args = {
					CaptionNormalMode = {
						order = 1,
						type = "description",
						name = L["Message Settings"],
					},
					ShowSpellEventsToggle = {
						order = 2,
						type = "toggle",
						width = "full",
						name = L["Report detected spell events"],
						desc = L["Report detected spell events TOOLTIP"],
						get = function() return self:OnGetShowSpellEventsToggle() end,
						set = function(_, value) self:OnSetShowSpellEventsToggle(value) end,
					},
					SpellSelect = {
						order = 3,
						type = "select",
						width = "full",
						name = L["Select a Spell"],
						desc = L["Select a Spell TOOLTIP"],
						values = self.RuntimeData.SpellList,
						get = function() return self:OnGetSpellSelect() end,
						set = function(_, value) self:OnSetSpellSelect(value) end,
					},
					NewSpellButton = {
						order = 4,
						type = "execute",
						name = L["New Spell"],
						desc = L["New Spell TOOLTIP"],
						func = function() self:OnClickNew() end,
					},
					DeleteSpellButton = {
						order = 5,
						type = "execute",
						name = L["Delete Spell"],
						desc = L["Delete Spell TOOLTIP"],
						func = function() self:OnClickDelete() end,
					},
					ChannelSelectCaption = {
						order = 6,
						type = "description",
						width= "full",
						name = L["Select the channels to use when announcing this spell when"],
					},
					ChannelSelectBG = {
						order = 7,
						type = "select",
						name = L["In a Battleground"],
						desc = L["In a Battleground TOOLTIP"],
						values = { -- color-coded to match default chat channel colors
							[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							--[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							--[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							--[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return self:OnGetChannelSelect("BG") end,
						set = function(_, value) self:OnSetChannelSelect("BG",value) end,
					},
					ChannelSelectRaid = {
						order = 8,
						type = "select",
						name = L["In a Raid"],
						desc = L["In a Raid TOOLTIP"],
						values = { -- color-coded to match default chat channel colors
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return self:OnGetChannelSelect("Raid") end,
						set = function(_, value) self:OnSetChannelSelect("Raid",value) end,
					},
					ChannelSelectParty = {
						order = 9,
						type = "select",
						name = L["In a Party"],
						desc = L["In a Party TOOLTIP"],
						values = { -- color-coded to match default chat channel colors
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							--[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return self:OnGetChannelSelect("Party") end,
						set = function(_, value) self:OnSetChannelSelect("Party",value) end,
					},
					ChannelSelectSolo = {
						order = 10,
						type = "select",
						name = L["By yourself"],
						desc = L["By yourself TOOLTIP"],
						values = { -- color-coded to match default chat channel colors
							--[L["BATTLEGROUND"]]	= "|cffff6b00"..L["BATTLEGROUND"],
							--[L["RAID"]]			= "|cffff6b00"..L["RAID"],
							--[L["PARTY"]]		= "|cffaaeeff"..L["PARTY"],
							[L["SAY"]]			=               L["SAY"],
							[L["EMOTE"]]		= "|cffff7b34"..L["EMOTE"],
							[L["YELL"]]			= "|cffff2f32"..L["YELL"],
							--[L["RAID_WARNING"]]	= "|cffffdbad"..L["RAID_WARNING"],
							[L["Silent"]]		=               L["Silent"], --tried light gray to indicate disabled, but it made the control appear to be disabled, so use the default color (white)
						},
						get = function() return self:OnGetChannelSelect("Solo") end,
						set = function(_, value) self:OnSetChannelSelect("Solo",value) end,
					},
					WhisperTargetToggle = {
						order = 11,
						type = "toggle",
						width = "full",
						name = L["Whisper the message to your target"],
						desc = L["Whisper the message to your target TOOLTIP"],
						get = function() return self:OnGetWhisperTarget() end,
						set = function(_, value) self:OnSetWhisperTarget(value) end,
					},
					FrequencySelect = {
						order = 12,
						type = "range",
						width = "full",
						name = L["How often do you want to announce this spell?"],
						desc = L["How often TOOLTIP"],
						min = 0.01, max = 1.00, step=0.01,
						isPercent = true,
						get = function() return self:OnGetFrequencySelect() end,
						set = function(_, value) self:OnSetFrequencySelect(value) end,
					},
					CooldownSlider = {
						order = 13,
						type = "range",
						width = "full",
						name = L["Cooldown on announcements for this spell"] ,
						desc = L["Cooldown TOOLTIP"],
						min =0, max = 600, step=1, -- allowing for 0 seconds up to 10 minutes
						get = function() return self:OnGetCooldownSlider() end,
						set = function(_, value) self:OnSetCooldownSlider(value) end,
					},
					-- special hidden text box for entering new spell names
					-- this is only shown after the user clicks on NEW
					NewSpellModeInstruct = {
						order = 1,
						type = "description",
						name = L["NEW SPELL INSTRUCTIONS"],
						guiHidden = true, --by default, until OnClickNew is called
					},
					NewSpellName = {
						order = 2,
						width = "full",
						type = "input",
						name = L["Enter a new Spell Name"],
						desc = L["Enter a new Spell Name TOOLTIP"],
						get = function() return "" end,
						set = function(_, value) self:OnSetNewSpellName(value) end,
						guiHidden = true, --by default, until OnClickNew is called
					},
					CancelNewSpellButton = {
						order = 3,
						type = "execute",
						name = L["Cancel New Spell"],
						desc = L["Cancel New Spell TOOLTIP"],
						func = function() self:OnClickCancelNewSpell() end,
						guiHidden = true, --by default, until OnClickNew is called
					},
					-- instructions will be shown to add a new spell if the list is empty
					AddFirstSpellInstruct = {
						order = 1,
						type = "description",
						name = L["ADD FIRST SPELL INSTRUCTIONS"],
						guiHidden = true, --by default, until OnClickNew is called
					},
					-- a list of up to 10 (MAX_MESSAGES_PER_SPELL) input boxes will be added by a loop below
					-- for entering/editing the random speeches for each spell
				}, --end args
			}, --end General
		}, --end args
	} -- end options
	
	-- add MAX_MESSAGES_PER_SPELL random speech input boxes
	for i=1,MAX_MESSAGES_PER_SPELL,1 do
		self.OptionsGUI.args.General.args["SpellMessage"..i] = {
			order = i + 14, -- should be the order of the last item listed above the message input boxes, plus i
			width = "full",
			type = "input",
			name = string.format(L["Random Speech i"], i), -- "Random Speech %d"
			desc = L["Random Speech i TOOLTIP"],
			get = function() return self:OnGetMessage(i) end,
			set = function(_, value) self:OnSetMessage(i, value) end,
			guiHidden = true, --by default, until ShowHideOptionsGUIControls is called
		}
	end
	
	-- add the options to the in-game interface options panel at Escape > Interface > Addons > SpeakinSpell
	-- The ordering here matters, it determines the order in the Blizzard Interface Options
	-- (or it would matter if we had sub-categories, which we currently don't)
	-- InterfaceOptions_AddCategory(self.OptionsGUI) -- done via Ace below

	LibStub("AceConfig-3.0"):RegisterOptionsTable("SpeakinSpell", self.OptionsGUI, nil)
	self.optionsFrames = {["General"] = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SpeakinSpell", L["SpeakinSpell"], nil, "General")}
	
	-- AddToBlizOptions returns the frame object.  Add a default button handler
	self.optionsFrames.General.default = function() self:ResetToDefaults() end
end



-------------------------------------------------------------------------------
-- GENERAL INITIALIZATION
-------------------------------------------------------------------------------


-- our saved data is stored in SpeakinSpellSavedData.SpellTable
-- we must declare the parent table SpeakinSpellSavedData
-- but we know whether real data has been loaded or not if SpeakinSpellSavedData.SpellTable exists
-- NOTE: SpeakinSpellSavedData.SpellTable may be an empty table
SpeakinSpellSavedData = {}

function SpeakinSpell:InitDefaultSavedData()
	self:DebugMsg("InitDefaultSavedData()")
	
	-- copy defaults into current settings using class-specific defaults
	local playerClass, englishClass = UnitClass("player"); --englishClass is the UPPERCASE class name in English
	SpeakinSpellSavedData.SpellTable = nil
	SpeakinSpellSavedData.SpellTable = {}
	SpeakinSpellSavedData.SpellTable = self:CopyTable( L.DEFAULT_DATA_PER_CLASS[englishClass] )
	SpeakinSpellSavedData.Version = SpeakinSpell.CURRENT_VERSION
	SpeakinSpellSavedData.ShowSpellEvents = false

	-- remove faction-specific spells
	for spellnamekey, SpellTableEntry in pairs(SpeakinSpellSavedData.SpellTable) do
		if SpellTableEntry.Faction ~= nil then
			if SpellTableEntry.Faction ~= UnitFactionGroup("player") then
				SpeakinSpellSavedData.SpellTable[spellnamekey] = nil
			end
		end
	end
	
	--self:DebugMsgDumpTable(SpeakinSpellSavedData.SpellTable, "SpeakinSpellSavedData.SpellTable")
	--self:DeleteInvalidSpells() -- we can't let this happen too soon during loading, or it will think all spells are invalid
	--self:DebugMsgDumpTable(SpeakinSpellSavedData.SpellTable, "SpeakinSpellSavedData.SpellTable")
end


function SpeakinSpell:ResetToDefaults()
	self:DebugMsg("ResetToDefaults()")
	
	SpeakinSpellSavedData.SpellTable = nil
	
	-- copy defaults into current settings using class-specific defaults
	self:InitDefaultSavedData()
	
	-- update/reset runtime data state
	self:InitRuntimeData()
	self:ShowHideOptionsGUIControls() -- this might be redundant but doesn't seem to hurt
	
	-- if the options GUI is showing right now, we need to refresh it to make it show the new data
	-- the only way I found to do this is to hide and then show the frame
	-- this occurs fast enough that I do not see a flicker when this happens
	if self.optionsFrames and self.optionsFrames.General and self.optionsFrames.General:IsVisible() then
		self.optionsFrames.General:Hide()
		self.optionsFrames.General:Show()
	end
	
	self:Print(L["Default settings restored"])
end


function SpeakinSpell:InitRuntimeData()
	self:DebugMsg("InitRuntimeData()")
	
	-- this is just a declaration
	self.RuntimeData = {
		CurrentSpellSelect = nil,	-- current spell selected in the options GUI
		SpellList = {},				-- complete list of spells displayed in the options GUI
		LastMessage = {},			-- just an empty table for now, this will be [spellname] = "last random speech triggered by this spell" to prevent repeats
		EnterNewSpellMode = false,	-- flag indicating that the user is entering a new spell for the list
	}
	
	for spellname,info in pairs(SpeakinSpellSavedData.SpellTable) do
		-- build the spell list for use by the options GUI
		self.RuntimeData.SpellList[spellname] = spellname
		self:DebugMsg("InitRuntimeData: added spell: " .. spellname)
		-- make sure we have a current spell selected for the options GUI
		if not self.RuntimeData.CurrentSpellSelected then
			self.RuntimeData.CurrentSpellSelected = spellname
		end
	end
	
	--self:DebugMsgDumpTable(self.RuntimeData.SpellList, "self.RuntimeData.SpellList", "")

	-- the options GUI may still have a reference to the old SavedData table.
	-- refresh it since we've rebuilt the table
	-- but if we're initializing then the OptionsGUI may not have been created yet
	if self.OptionsGUI then
		self.OptionsGUI.args.General.args.SpellSelect.values = self.RuntimeData.SpellList
	end
end


function SpeakinSpell:OnInitialize()
	self:DebugMsg("OnInitialize()")
    -- Called when the addon is loaded

	self:RegisterAllEvents()
	self:RegisterChatCommands()	-- register chat commands
	-- the rest of the initialization does not occur until VARIABLES_LOADED
	
	-- done loading
    self:Print( string.format(L["Loaded version %s"], self.CURRENT_VERSION) )
end


-------------------------------------------------------------------------------
-- EVENT HANDLERS
-------------------------------------------------------------------------------


function SpeakinSpell:FixOldSavedData()
	-- prior to v3.0.3.08, a version number was not saved in the SpeakinSpellSavedData
	-- create an older version number
	if not SpeakinSpellSavedData.Version then
		SpeakinSpellSavedData.Version = "3.0.3.07"
	end
	
	--[[ example of how this can work in the future - this is not actually necessary for 3.0.3.08
	if SpeakinSpellSavedData.Version < "3.0.3.08" then
		-- make changes to SpeakinSpellSavedData, updating from older versions to 3.0.3.08
		
		-- the data has been brought up to date with 3.0.3.08
		SpeakinSpellSavedData.Version = "3.0.3.08"
	end

	if SpeakinSpellSavedData.Version < "3.0.3.09" then
		-- make changes to SpeakinSpellSavedData
		-- data is guaranteed to be in 3.0.3.08 format now, due to the previous block
		
		-- the data has been brought up to date with 3.0.3.09
		SpeakinSpellSavedData.Version = "3.0.3.09"
	end
	--]]
	
	-- saved data is now up to date with the latest version
	SpeakinSpellSavedData.Version = SpeakinSpell.CURRENT_VERSION
end


function SpeakinSpell:VARIABLES_LOADED()
	self:DebugMsg("VARIABLES_LOADED()")
	
	-- SpeakinSpellSavedData will always exist
	-- SpeakinSpellSavedData.SpellTable only exists if actual data has been loaded
	-- NOTE: SpeakinSpellSavedData.SpellTable may be empty, but its parent table SpeakinSpellSavedData never will be
	if SpeakinSpellSavedData.SpellTable then
		-- If the format of the saved data has changed due to upgrading to a newer version
		-- then fix the old saved data for compatibility
		self:FixOldSavedData()
	else
		self:DebugMsg("VARIABLES_LOADED() - SpeakinSpellSavedData.SpellTable does not exist, using defaults")
		-- new user: load defaults
		self:InitDefaultSavedData()
	end
	self:DebugMsgDumpTable(SpeakinSpellSavedData.SpellTable, "SpeakinSpellSavedData.SpellTable")

	-- set up runtime data that we don't save, only after the saved variables are loaded
	self:InitRuntimeData()
	
	-- set up the Options GUI config - we need the Runtime Data before we can do this
	self:SetupOptionsGUI()
end


function SpeakinSpell:SPELL_AURA_APPLIED(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, spellId, spellName, spellSchool, auraType)
	-- we're only concerned with buffs
	if auraType ~= "BUFF" then return end
	-- we're only concerned with buffs we cast on ourselves
	if dstName ~= srcName then return end
	if srcName ~= UnitName("player") then return end
	-- ignore redundant events that are also applied by the last UNIT_SPELLCAST_SENT event
	if spellName == self.RuntimeData.Last_UNIT_SPELLCAST_SENT then return end
	
	-- process the spell
	-- this function is shared with the UNIT_SPELLCAST_SENT event handler
	self:OnValidSpellcastDetected(spellName, "", UnitName("player")) -- the target is yourself and there's no rank
end


function SpeakinSpell:COMBAT_LOG_EVENT_UNFILTERED(_,timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	--self:Print("COMBAT_LOG_EVENT_UNFILTERED: "..eventtype)
	if eventtype == "SPELL_AURA_APPLIED" then
		self:SPELL_AURA_APPLIED(timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	end
end


function SpeakinSpell:UNIT_SPELLCAST_SENT(event, caster, name, rank, target)
	self:DebugMsg("UNIT_SPELLCAST_SENT()")

	--[[	
	self:DebugMsgDumpString("event",	event)
	self:DebugMsgDumpString("caster",	caster)
	self:DebugMsgDumpString("name",		name)
	self:DebugMsgDumpString("rank",		rank)
	self:DebugMsgDumpString("target",	target)
	--]]
	
	-- only speak for spells cast by the player, not anyone else
	if caster ~= "player" then
		self:DebugMsg("UNIT_SPELLCAST_SENT - do not speak for spells cast by: " .. caster)
		return
	end
	
	-- prevent SPELL_AURA_APPLIED from triggering a redundant event for a matching aura cast by this spell
	self.RuntimeData.Last_UNIT_SPELLCAST_SENT = name
	
	-- process the spell
	-- this function is shared with the SPELL_AURA_APPLIED event handler
	self:OnValidSpellcastDetected(name, rank, target)
end


function SpeakinSpell:OnValidSpellcastDetected(name, rank, target)
	self:DebugMsg("OnValidSpellcastDetected()")
	
	if SpeakinSpellSavedData.ShowSpellEvents then
		if rank ~= "" then
			self:Print(name.." ("..rank..")")
		else
			self:Print(name)
		end
	end
	
	local SpellTableEntry, spellnamekey = self:GetSpellTableEntry(name, rank)
	if not SpellTableEntry then
		self:DebugMsg("OnValidSpellcastDetected - no messages configured for: " .. name)
		return
	end
	
	-- check the frequency setting against a random number to determine if we should speak or not
	-- NOTE: if we are running on saved data from an older version of SpeakinSpell
	--		it is possible that the Frequency value is nil
	--		in that case, use the default value of 100%
	if SpellTableEntry.Frequency == nil then
		SpellTableEntry.Frequency = FREQUENCY_DEFAULT
	end
	local Chance = math.random(1,100)
	local Frequency = SpellTableEntry.Frequency*100
	if Chance > Frequency then
		return
	end
	
	-- check for a cooldown limit on announcing this spell
	-- NOTE: if we're dealing with older saved data, then the Cooldown might not be defined
	--		make sure it exists
	if SpellTableEntry.Cooldown == nil then
		SpellTableEntry.Cooldown = COOLDOWN_DEFAULT
	end
	-- NOTE: if we have not spoken for this spell yet, then
	--		self.RuntimeData.LastMessageTime or self.RuntimeData.LastMessageTime[spellname] might not exist
	--		in that case, set elapsed = cooldown, so that it will seem like enough time has passed
	local now = GetTime()
	if self.RuntimeData.LastMessageTime == nil then
		self.RuntimeData.LastMessageTime = {}
	end
	local last = self.RuntimeData.LastMessageTime[spellnamekey]
	if last == nil then
		last = 0
	end
	local elapsed = SpellTableEntry.Cooldown
	if last then
		elapsed = now - last
	end
	self:DebugMsg( string.format("%s: now:%d, last:%d, elapsed:%d ,CD:%d", name, now, last, elapsed, SpellTableEntry.Cooldown) )
	-- Now finally actually check if the cooldown is in effect
	if elapsed < SpellTableEntry.Cooldown then
		-- the cooldown is in effect, so be silent
		return
	end
	
	-- speak a random message
	self:SpeakForSpell(SpellTableEntry, spellnamekey, name, target)
end


function SpeakinSpell:RegisterAllEvents()
	self:DebugMsg("RegisterAllEvents()")
	-- register for spellcasting events
	-- which is our hook too know when to speak for a spell
	self:RegisterEvent("UNIT_SPELLCAST_SENT")
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self:RegisterEvent("VARIABLES_LOADED")
end


function SpeakinSpell:OnEnable()
	self:DebugMsg("OnEnable()")
	-- Called when the addon is enabled
	-- self:RegisterAllEvents() -- did this OnInitialize
	self:ShowHideOptionsGUIControls() -- make sure the options GUI is setup correctly
end


function SpeakinSpell:OnDisable()
	self:DebugMsg("OnDisable()")
    -- Called when the addon is disabled
	-- we don't need to do anything here
end


-------------------------------------------------------------------------------
-- SLASH COMMAND HANDLERS
-------------------------------------------------------------------------------


function SpeakinSpell:RegisterChatCommands()
	self:DebugMsg("RegisterChatCommands()")
	
	-- NOTE: support both localized and default English slash commands
	local SlashCommands = {
		L["/SpeakinSpell"],
		  "/SpeakinSpell",
		L["/speakinspell"],
		  "/speakinspell",
		L["/SS"],
		  "/SS",
		L["/ss"],
		  "/ss",
	}
	
	for k,v in pairs(SlashCommands) do
		self:RegisterChatCommand(v,"OnSlashCommand")
	end
end


function SpeakinSpell:OnSlashCommand(input)
	self:DebugMsg("OnSlashCommand()")
	self:DebugMsgDumpString("input", input)

	-- table of supporting slash commands
	-- NOTE: support both English and current locale
	--	non-English game clients can typically use both translated and untranslated slash commands
	--	for built-in functions like /bg chat, etc, so it makes sense here as well
	local SlashCommandTable = {
		-- /ss options	- also show options
		[L["/options"]] = function() self:ShowOptions() end,
		[   "options" ] = function() self:ShowOptions() end,
		
		-- /ss reset	- reset to defaults
		[L["/reset"]] = function() self:ResetToDefaults() end,
		[   "reset" ] = function() self:ResetToDefaults() end,
		
		-- /ss advertise- show an advertisement for SpeakinSpell
		[L["/advertise"]] = function() self:Advertise() end,
		[   "advertise" ] = function() self:Advertise() end,
		
		-- /ss url- show the url
		[L["/url"]] = function() self:Print(URL) end,
		[   "url" ] = function() self:Print(URL) end,
	}
	
	input = string.lower(input)
	-- the default command is to show the options
	if input == nil or input == "" then
		input = "options"
	end
	
	-- what did the user enter?
	Func = SlashCommandTable[input]
	
	-- process the command
	if Func then
		Func()
	else -- in all other cases, show help
		self:Print(L["/help TEXT"])
	end
end


function SpeakinSpell:ShowOptions()
	self:DebugMsg("ShowOptions()")
	--self:DebugMsgDumpTable(self.RuntimeData.SpellList, "self.RuntimeData.SpellList", "")

	-- this is probably not necessary, but make sure this is up to date before opening the options panel
	self:ShowHideOptionsGUIControls()
	
	-- open interface options to show SpeakinSpell
	InterfaceOptionsFrame_OpenToCategory(L["SpeakinSpell"])
end


function SpeakinSpell:Advertise()
	-- advertise into a channel for the kind of group you are in
	local Channel = "SAY"
	if (GetNumBattlefieldScores() > 0) then
		Channel = "SAY" -- don't advertise in BG chat, that would be annoying
	elseif (GetNumRaidMembers() > 0) then
		Channel = "RAID"
	elseif (GetNumPartyMembers() > 0) then
		Channel = "PARTY"
	else -- solo
		Channel = "SAY"
	end
	
	SendChatMessage(L["ADVERTISEMENT"], Channel)
	SendChatMessage(URL, Channel)	
end


-------------------------------------------------------------------------------
-- RANDOM SPEECHES - THE CORE FUNCTIONALITY
-------------------------------------------------------------------------------


function SpeakinSpell:GetSpellTableEntry(name, rank)

	namekey = string.lower(name) -- force to lowercase for case-insensitive string matching
	rankkey = string.lower(rank) -- force to lowercase for case-insensitive string matching
	
	-- check for an entry for the name with rank
	local spellnamekey = namekey .. " (" .. rankkey .. ")"
	local SpellTableEntry = SpeakinSpellSavedData.SpellTable[ spellnamekey ]
	if SpellTableEntry then
		return SpellTableEntry, spellnamekey
	end
	
	-- try the name only
	SpellTableEntry = SpeakinSpellSavedData.SpellTable[ namekey ]
	if SpellTableEntry then
		return SpellTableEntry, namekey
	end
	
	-- no spell table entry defined for this name and rank
	return nil,nil
end
	
	
function SpeakinSpell:SelectRandomMessage(spellnamekey, SpellTableEntry)
	self:DebugMsg("SelectRandomMessage( "..spellnamekey.." )")
	assert(SpellTableEntry and SpellTableEntry.Messages) -- this function should not be called unless we have a valid table for this spell
	
	-- select a random message from the user-specific table
	
	local sizelist = #(SpellTableEntry.Messages)
	local tempnum = math.random(1, sizelist);
	
	msg = SpellTableEntry.Messages[tempnum]
	assert(msg)
	assert(type(msg) == "string")
	
	-- avoid repeating the same message we used last time
	-- unless there's only one in the list
	while sizelist > 1 and msg == self.RuntimeData.LastMessage[spellnamekey] do
		tempnum = math.random(1, sizelist);
		msg = SpellTableEntry.Messages[tempnum]
	end
	
	-- remember we selected this, so we don't repeat it next time
	self.RuntimeData.LastMessage[spellnamekey] = msg
	
	return msg
end


function SpeakinSpell:GetChatChannelForSpell(SpellTableEntry)
	self:DebugMsg("GetChatChannelForSpell()")

	-- NOTE: the SpellTableEntry.Channels table stores channel names that are compatible with SendChatMessage
	--	see also OnSetChannelSelect / OnGetChannelSelect
	if (GetNumBattlefieldScores() > 0) then
		return SpellTableEntry.Channels.BG
	elseif (GetNumRaidMembers() > 0) then
		return SpellTableEntry.Channels.Raid
	elseif (GetNumPartyMembers() > 0) then
		return SpellTableEntry.Channels.Party
	else -- solo
		return SpellTableEntry.Channels.Solo
	end
end


function SpeakinSpell:SpeakForSpell(SpellTableEntry, spellnamekey, spellname, target)
	-- determine which channel we will speak in
	channel = self:GetChatChannelForSpell(SpellTableEntry)
	local WhisperTarget = SpellTableEntry.WhisperTarget and target and (target ~= "") and (target ~= "Unknown") and (target ~= UnitName("player")) and UnitIsFriend("player","target")
	if not channel and not WhisperTarget then
		-- this spell's speaches have been silenced for the current situation
		return
	end
	
	msg = self:SelectRandomMessage(spellnamekey, SpellTableEntry)
	
	local targetclass = L["<no target class>"]
	local targetrace = L["<no target race>"]
	
	if not target or target == "" or target == "Unknown" then
		target = L["<no target>"]
	else
		targetclass = UnitClass("target")
		targetrace = UnitRace("target")
	end
	
	local petname = UnitName("pet")
	if petname == nil or petname == "" then
		petname = L["<no pet>"]
	end
		
	-- NOTE: also support the english variable names if running in a non-english locale
	--		this is meant to help support the fact that documentation in-game is lacking for this variable list
	local Substitutions = {
		[L["<player>"]]		= UnitName("player"),	-- the player's name / your name
		[  "<player>" ]		= UnitName("player"),	-- the player's name / your name
		
		[L["<target>"]]		= target, -- the target's name (safety check for nil has been done above)
		[  "<target>" ]		= target, -- the target's name (safety check for nil has been done above)
		
		[L["<targetclass>"]]= targetclass, -- the target's name (safety check for nil has been done above)
		[  "<targetclass>" ]= targetclass, -- the target's name (safety check for nil has been done above)
		
		[L["<targetrace>"]]	= targetrace, -- the target's name (safety check for nil has been done above)
		[  "<targetrace>" ]	= targetrace, -- the target's name (safety check for nil has been done above)
		
		[L["<spellname>"]]	= spellname,
		[  "<spellname>" ]	= spellname,
		
		[L["<pet>"]]		= petname,
		[  "<pet>" ]		= petname,
		
		-- auto-substitutions like MS Word would use
		[L["<TM>"]]			= "™",	-- Trademark symbol
		[  "<TM>" ]			= "™",	-- Trademark symbol
		
		[L["<R>"]]			= "®",	-- Registered Trademark symbol
		[  "<R>" ]			= "®",	-- Registered Trademark symbol
		
		[L["<C>"]]			= "©",	-- Copyright (C)
		[  "<C>" ]			= "©",	-- Copyright (C)
	}
	
	-- insert variable values into the message text
	for key,value in pairs(Substitutions) do
		msg = string.gsub(msg, key, value);
	end
	
	-- record the time when we spoke for this spell
	self.RuntimeData.LastMessageTime[spellnamekey] = GetTime()
	
	-- say the message to the chat, if applicable
	if channel then
		SendChatMessage(msg, channel);
	end
	
	-- send the message to a whisper to the target, if applicable
	if WhisperTarget then
		SendChatMessage(msg, "WHISPER", nil, target);
	end
end


